@@ -22,20 +22,23 @@ import gtk
 from debug import *
 
 _ = lambda x: x
+N_ = lambda x: x
 def set_gettext_function (fn):
     global _
     _ = fn
 
 class AuthDialog(gtk.Dialog):
-    AUTH_FIELD={'username': _("Username:"),
-                'password': _("Password:"),
-                'domain': _("Domain:")}
+    AUTH_FIELD={'username': N_("Username:"),
+                'password': N_("Password:"),
+                'domain': N_("Domain:")}
 
-    def __init__ (self, title=_("Authentication"), parent=None,
+    def __init__ (self, title=None, parent=None,
                   flags=gtk.DIALOG_MODAL | gtk.DIALOG_NO_SEPARATOR,
                   buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                            gtk.STOCK_OK, gtk.RESPONSE_OK),
                   auth_info_required=['username', 'password']):
+        if title == None:
+            title = _("Authentication")
         gtk.Dialog.__init__ (self, title, parent, flags, buttons)
         self.auth_info_required = auth_info_required
         self.set_default_response (gtk.RESPONSE_OK)
@@ -60,7 +63,7 @@ class AuthDialog(gtk.Dialog):
         self.field_entry = []
         for i in range (num_fields):
             field = auth_info_required[i]
-            label = gtk.Label (self.AUTH_FIELD.get (field, field))
+            label = gtk.Label (_(self.AUTH_FIELD.get (field, field)))
             label.set_alignment (0, 0.5)
             table.attach (label, 0, 1, i, i + 1)
             entry = gtk.Entry ()
@@ -154,6 +157,8 @@ class Connection:
                 fn = getattr (self._connection, fname)
                 c = self._connection
 
+            cups.setUser (self._use_user)
+
             try:
                 result = fn.__call__ (*args, **kwds)
 
@@ -166,8 +171,11 @@ class Connection:
             except cups.IPPError, (e, m):
                 if not self._cancel and (e == cups.IPP_NOT_AUTHORIZED or
                                          e == cups.IPP_FORBIDDEN):
-                    self._failed ()
+                    self._failed (e == cups.IPP_FORBIDDEN)
                 else:
+                    if self._cancel and not self._cannot_auth:
+                        raise cups.IPPError (0, _("Operation canceled"))
+
                     raise
             except cups.HTTPError, (s,):
                 if not self._cancel and (s == cups.HTTP_UNAUTHORIZED or
@@ -202,11 +210,13 @@ class Connection:
             self._forbidden = False
             self._auth_called = False
             self._cancel = False
+            self._cannot_auth = False
             self._dialog_shown = False
             cups.setPasswordCB (self._password_callback)
             debugprint ("Authentication: password callback set")
             return 1
 
+        debugprint ("Forbidden: %s" % self._forbidden)
         if not self._has_failed:
             # Tried the operation and it worked.  Return 0 to signal to
             # break out of the loop.
@@ -216,7 +226,7 @@ class Connection:
         # Reset failure flag.
         self._has_failed = False
 
-        if self._passes == 2:
+        if self._passes >= 2:
             # Tried the operation without a password and it failed.
             if (self._try_as_root and
                 self._user != 'root' and
@@ -244,6 +254,7 @@ class Connection:
             # We aren't even getting a chance to supply credentials.
             debugprint ("Authentication: giving up")
             self._cancel = True
+            self._cannot_auth = True
             return 1
 
         # Reset the flag indicating whether we were given an auth callback.
@@ -268,8 +279,15 @@ class Connection:
         d.set_prompt (self._prompt)
         d.set_auth_info ([self._use_user, ''])
         d.field_grab_focus ('password')
+        d.set_keep_above (True)
+        d.show_all ()
+        d.show_now ()
+        gtk.gdk.keyboard_grab (d.window, True)
+        gtk.gdk.pointer_grab (d.window, True)
         self._dialog_shown = True
         response = d.run ()
+        gtk.gdk.pointer_ungrab ()
+        gtk.gdk.keyboard_ungrab ()
         (self._use_user,
          self._use_password) = d.get_auth_info ()
         d.destroy ()
