@@ -38,11 +38,11 @@ class Printer:
         self.connection = connection
         self.class_members = []
         self.update (**kw)
+        self._ppd = None # load on demand
 
-        if self.is_class:
-            self._ppd = False
-        else:
-            self._ppd = None # load on demand
+    def __del__ (self):
+        if self._ppd != None:
+            os.unlink(self._ppd)
 
     def __repr__ (self):
         return "<cupshelpers.Printer \"%s\">" % self.name
@@ -184,17 +184,21 @@ class Printer:
         @returns: cups.PPD object, or False for raw queues
         @raise cups.IPPError: IPP error
         """
+        result = None
         if self._ppd is None:
             try:
-                filename = self.connection.getPPD(self.name)
-                self._ppd = cups.PPD(filename)
-                os.unlink(filename)
+                self._ppd = self.connection.getPPD(self.name)
+                result = cups.PPD (self._ppd)
             except cups.IPPError, (e, m):
                 if e == cups.IPP_NOT_FOUND:
-                    self._ppd = False
+                    result = False
                 else:
                     raise
-        return self._ppd
+
+        if result != False and self._ppd != None:
+            result = cups.PPD (self._ppd)
+
+        return result
 
     def setOption(self, name, value):
         """
@@ -377,7 +381,7 @@ class Printer:
                 pass
 
             if s == cups.HTTP_NOT_FOUND:
-                return
+                return False
 
             raise cups.HTTPError (s)
 
@@ -400,16 +404,15 @@ class Printer:
                 self.connection.putFile (resource, tmpfname)
             except cups.HTTPError, (s,):
                 os.remove (tmpfname)
-                return
-
-            # Now reconnect because the server needs to reload.
-            self.reconnect ()
+                return False
 
         try:
             os.remove (tmpfname)
         except OSError:
             pass
 
+        return changed
+
 def getPrinters(connection):
     """
     Obtain a list of printers.
@@ -514,7 +517,7 @@ class Device:
                 return -1
             return 1
         if not self.is_class and (self.type != other.type):
-            # "hp"/"hpfax" before * before "usb" before "parallel" before
+            # "hp"/"hpfax" before "usb" before * before "parallel" before
             # "serial"
             if other.type == "serial":
                 return -1
@@ -524,14 +527,14 @@ class Device:
                 return -1
             if self.type == "parallel":
                 return 1
-            if other.type == "usb":
-                return -1
-            if self.type == "usb":
-                return 1
             if other.type == "hp" or other.type == "hpfax":
                 return 1
             if self.type == "hp" or self.type == "hpfax":
                 return -1
+            if other.type == "usb":
+                return 1
+            if self.type == "usb":
+                return -1
         result = cmp(bool(self.id), bool(other.id))
         if not result:
             result = cmp(self.info, other.info)
@@ -558,7 +561,11 @@ class _PrintersConf:
                 self.connection.getFile(file, filename)
         except cups.HTTPError, e:
             os.close(fd)
-            os.unlink(filename)
+            try:
+                os.unlink(filename)
+            except OSError:
+                pass
+
             if (e.args[0] == cups.HTTP_UNAUTHORIZED or
                 e.args[0] == cups.HTTP_NOT_FOUND):
                 return []
@@ -784,9 +791,12 @@ def missingPackagesAndExecutables(ppd):
         # Look for '*cupsFilter' lines in the PPD and check that
         # the filters are installed.
         (tmpfd, tmpfname) = tempfile.mkstemp ()
+        os.unlink (tmpfname)
         ppd.writeFd (tmpfd)
+        os.lseek (tmpfd, 0, os.SEEK_SET)
+        f = os.fdopen (tmpfd, "r")
         search = "*cupsFilter:"
-        for line in file (tmpfname).readlines ():
+        for line in f.readlines ():
             if line.startswith (search):
                 line = line[len (search):].strip ().strip ('"')
                 try:
