@@ -49,7 +49,7 @@ from statereason import StateReason
 statereason.set_gettext_function (_)
 errordialogs.set_gettext_function (_)
 
-pkgdata = config.Paths ().get_path ('pkgdatadir')
+pkgdata = config.pkgdatadir
 GLADE="applet.glade"
 ICON="printer"
 SEARCHING_ICON="document-print-preview"
@@ -296,6 +296,7 @@ class JobViewer (monitor.Watcher):
         if bus == None:
             bus = dbus.SystemBus ()
 
+        self.set_process_pending (True)
         self.host = cups.getServer ()
         self.port = cups.getPort ()
         self.encryption = cups.getEncryption ()
@@ -309,9 +310,22 @@ class JobViewer (monitor.Watcher):
 
     def cleanup (self):
         self.monitor.cleanup ()
+
+        # Close any open notifications.
+        for l in [self.new_printer_notifications.values (),
+                  self.auth_notifications.values (),
+                  self.state_reason_notifications.values ()]:
+            for notification in l:
+                if notification.get_data ('closed') != True:
+                    notification.close ()
+                    notification.set_data ('closed', True)
+
         if self.exit_handler:
             self.exit_handler (self)
 
+    def set_process_pending (self, whether):
+        self.process_pending_events = whether
+
     # Handle "special" status icon
     def set_special_statusicon (self, iconname):
         self.special_status_icon = True
@@ -328,9 +342,6 @@ class JobViewer (monitor.Watcher):
         notification.set_data ('printer-name', printer)
         notification.connect ('closed', self.on_new_printer_notification_closed)
         self.set_statusicon_visibility ()
-        # Let the icon show itself, ready for the notification
-        while gtk.events_pending ():
-            gtk.main_iteration ()
         notification.attach_to_status_icon (self.statusicon)
         notification.show ()
 
@@ -460,18 +471,58 @@ class JobViewer (monitor.Watcher):
     def on_troubleshoot_quit(self, troubleshooter):
         del self.troubleshooter
 
-    def add_job (self, job, data):
+    def add_job (self, job, data, connection=None):
         store = self.store
         iter = self.store.append (None)
         store.set_value (iter, 0, job)
         store.set_value (iter, 1, data.get('job-originating-user-name',
                                            _("Unknown")))
         store.set_value (iter, 2, data.get('job-name', _("Unknown")))
+        debugprint ("Job %d added" % job)
         self.jobiters[job] = iter
-        self.update_job (job, data)
-        store.set_value (iter, 5, _("a minute ago"))
+        if not self.job_creation_times_timer:
+            t = gobject.timeout_add (1000, self.update_job_creation_times)
+            self.job_creation_times_timer = t
+
+        self.update_job (job, data, connection=connection)
+
+    def update_job (self, job, data, connection=None):
+        # Fetch required attributes for this job if they are missing.
+        r = self.required_job_attributes - set (data.keys ())
+        if (data.get ('job-state', cups.IPP_JOB_HELD) != cups.IPP_JOB_HELD):
+            r -= set (['job-hold-until'])
+
+        if r:
+            attrs = None
+            try:
+                if connection == None:
+                    try:
+                        connection = cups.Connection (host=self.host,
+                                                      port=self.port,
+                                                      encryption=self.encryption)
+                    except TypeError:
+                        # Requires pycups >= 1.9.40.
+                        cups.setServer (self.host)
+                        cups.setPort (self.port)
+                        cups.setEncryption (self.encryption)
+                        connection = cups.Connection ()
+
+                debugprint ("requesting %s" % r)
+                r = list (r)
+                try:
+                    attrs = connection.getJobAttributes (job,
+                                                         requested_attributes=r)
+                except TypeError:
+                    # requested_attributes requires pycups >= 1.9.42
+                    attrs = connection.getJobAttributes (jobid)
+            except RuntimeError:
+                pass
+            except AttributeError:
+                pass
+
+            if attrs:
+                data.update (attrs)
 
-    def update_job (self, job, data):
         store = self.store
         iter = self.jobiters[job]
         self.jobs[job] = data
@@ -532,19 +583,29 @@ class JobViewer (monitor.Watcher):
                 notification.set_timeout (pynotify.EXPIRES_NEVER)
                 notification.connect ('closed',
                                       self.on_auth_notification_closed)
-                self.set_statusicon_visibility ()
-                notification.attach_to_status_icon (self.statusicon)
                 notification.add_action ("authenticate", _("Authenticate"),
                                          self.on_auth_notification_authenticate)
-                notification.show ()
                 self.auth_notifications[job] = notification
+                debugprint ("auth notification opened for job %s" % job)
+                self.set_statusicon_visibility ()
+
+                # In set_statusicon_visibility we process pending
+                # events, so we need to check that we still have a
+                # notification to show.
+                if notification.get_data ('closed') != True:
+                    notification.attach_to_status_icon (self.statusicon)
+                    notification.show ()
             elif (not job_requires_auth and
                   self.auth_notifications.has_key (job)):
+                debugprint ("job %s no longer requires auth" % job)
                 self.auth_notifications[job].close ()
+                self.auth_notifications[job].set_data ('closed', True)
+                del self.auth_notifications[job]
 
     def on_auth_notification_closed (self, notification, reason=None):
         job = notification.get_data ('job-id')
         debugprint ("auth notification closed for job %s" % job)
+        self.auth_notifications[job].set_data ('closed', True)
         del self.auth_notifications[job]
 
     def on_auth_notification_authenticate (self, notification, action):
@@ -626,15 +687,21 @@ class JobViewer (monitor.Watcher):
         dialog.connect ('delete-event', self.auth_info_dialog_delete)
         dialog.set_data ('job-id', job)
         dialog.show_all ()
+        dialog.set_keep_above (True)
+        dialog.show_now ()
+        gtk.gdk.keyboard_grab (dialog.window, True)
+        gtk.gdk.pointer_grab (dialog.window, True)
 
     def auth_info_dialog_delete (self, dialog, event):
         self.auth_info_dialog_response (dialog, gtk.RESPONSE_CANCEL)
 
     def auth_info_dialog_response (self, dialog, response):
-        dialog.hide ()
+        gtk.gdk.pointer_ungrab ()
+        gtk.gdk.keyboard_ungrab ()
         jobid = dialog.get_data ('job-id')
         del self.auth_info_dialogs[jobid]
         if response != gtk.RESPONSE_OK:
+            dialog.destroy ()
             return
 
         auth_info = dialog.get_auth_info ()
@@ -650,6 +717,8 @@ class JobViewer (monitor.Watcher):
             self.show_IPP_Error (e, m)
             pass
 
+        dialog.destroy ()
+
     def set_statusicon_visibility (self):
         if not self.trayicon:
             return
@@ -661,7 +730,10 @@ class JobViewer (monitor.Watcher):
             return
 
         open_notifications = len (self.new_printer_notifications.keys ())
-        open_notifications += len (self.state_reason_notifications.keys ())
+        open_notifications += len (self.auth_notifications.keys ())
+        for reason, notification in self.state_reason_notifications.iteritems():
+            if notification.get_data ('closed') != True:
+                open_notifications += 1
         num_jobs = len (self.jobs.keys ())
 
         debugprint ("open notifications: %d" % open_notifications)
@@ -672,6 +744,10 @@ class JobViewer (monitor.Watcher):
                                      open_notifications > 0 or
                                      num_jobs > self.num_jobs_when_hidden)
 
+        # Let the icon show/hide itself before continuing.
+        while self.process_pending_events and gtk.events_pending ():
+            gtk.main_iteration ()
+
     def on_treeview_popup_menu (self, treeview):
         event = gtk.gdk.Event (gtk.gdk.NOTHING)
         self.show_treeview_popup_menu (treeview, event, 0)
@@ -943,32 +1019,25 @@ class JobViewer (monitor.Watcher):
         (title, text) = reason.get_description ()
         notification = pynotify.Notification (title, text, 'printer')
         reason.user_notified = True
-        notification.set_data ('printer-state-reason', reason)
         notification.set_urgency (urgency)
         notification.set_timeout (pynotify.EXPIRES_NEVER)
         notification.connect ('closed',
                               self.on_state_reason_notification_closed)
         self.state_reason_notifications[reason.get_tuple ()] = notification
         self.set_statusicon_visibility ()
-        # Let the icon show itself, ready for the notification
-        while gtk.events_pending ():
-            gtk.main_iteration ()
         notification.attach_to_status_icon (self.statusicon)
         notification.show ()
 
     def on_state_reason_notification_closed (self, notification, reason=None):
         debugprint ("Notification %s closed" % repr (notification))
-        reason = notification.get_data ('printer-state-reason')
-        tuple = reason.get_tuple ()
-        if self.state_reason_notifications[tuple] == notification:
-            del self.state_reason_notifications[tuple]
-            self.set_statusicon_visibility ()
-            return
-
-        debugprint ("Unable to find closed notification")
+        notification.set_data ('closed', True)
+        self.set_statusicon_visibility ()
+        return
 
     ## monitor.Watcher interface
     def current_printers_and_jobs (self, mon, printers, jobs):
+        monitor.Watcher.current_printers_and_jobs (self, mon, printers, jobs)
+        self.set_process_pending (False)
         self.store.clear ()
         self.jobs = {}
         self.jobiters = {}
@@ -983,47 +1052,7 @@ class JobViewer (monitor.Watcher):
                 printer = uri
             jobdata['job-printer-name'] = printer
 
-            self.add_job (jobid, jobdata)
-
-            # Fetch required attributes for these jobs.
-            attrs = None
-            r = self.required_job_attributes - set (jobdata.keys ())
-            if (jobdata.get ('job-state', cups.IPP_JOB_HELD) !=
-                cups.IPP_JOB_HELD):
-                r -= set (['job-hold-until'])
-
-            if not r:
-                debugprint ("no further attributes required")
-                continue
-
-            try:
-                if connection == None:
-                    try:
-                        connection = cups.Connection (host=self.host,
-                                                      port=self.port,
-                                                      encryption=self.encryption)
-                    except TypeError:
-                        # Requires pycups >= 1.9.40.
-                        cups.setServer (self.host)
-                        cups.setPort (self.port)
-                        cups.setEncryption (self.encryption)
-                        connection = cups.Connection ()
-
-                try:
-                    debugprint ("requesting %s" % r)
-                    attrs = connection.getJobAttributes (jobid,
-                                                         requested_attributes=r)
-                except TypeError:
-                    # requested_attributes requires pycups >= 1.9.42
-                    attrs = connection.getJobAttributes (jobid)
-            except RuntimeError:
-                pass
-            except AttributeError:
-                pass
-
-            if attrs:
-                jobdata.update (attrs)
-                self.update_job (jobid, jobdata)
+            self.add_job (jobid, jobdata, connection=connection)
 
         self.jobs = jobs
         self.active_jobs = set()
@@ -1031,6 +1060,7 @@ class JobViewer (monitor.Watcher):
             if self.job_is_active (jobdata):
                 self.active_jobs.add (jobid)
 
+        self.set_process_pending (True)
         self.update_status ()
 
     def job_added (self, mon, jobid, eventname, event, jobdata):
@@ -1073,6 +1103,12 @@ class JobViewer (monitor.Watcher):
         elif jobid in self.active_jobs:
             self.active_jobs.remove (jobid)
 
+        if (jobdata.has_key ('job-hold-until') and
+            not event.has_key ('job-hold-until')):
+            del jobdata['job-hold-until']
+        self.update_job (jobid, jobdata)
+        jobdata = self.jobs[jobid]
+
         # Look out for stopped jobs.
         if (self.trayicon and eventname == 'job-stopped' and
             not jobid in self.stopped_job_prompts):
@@ -1080,45 +1116,12 @@ class JobViewer (monitor.Watcher):
             # of some sort, or it might be that the backend requires
             # authentication.  If the latter, the job will be held not
             # stopped, and the job-hold-until attribute will be
-            # 'auth-info-required'.  This will be checked for in
+            # 'auth-info-required'.  This was already checked for in
             # update_job.
-            if jobdata['job-state'] == cups.IPP_JOB_HELD:
-                r = self.required_job_attributes - set (jobdata.keys ())
-                if not r:
-                    debugprint ("no further attributes required")
-                else:
-                    try:
-                        # Fetch the job-hold-until attribute, as this is
-                        # not provided in the notification attributes.
-                        try:
-                            c = cups.Connection (host=self.host,
-                                                 port=self.port,
-                                                 encryption=self.encryption)
-                        except TypeError:
-                            # Requires pycups >= 1.9.40.
-                            cups.setServer (self.host)
-                            cups.setPort (self.port)
-                            cups.setEncryption (self.encryption)
-                            c = cups.Connection ()
-
-                        try:
-                            debugprint ("requesting %s" % r)
-                            attrs = c.getJobAttributes (jobid,
-                                                        requested_attributes=r)
-                        except TypeError:
-                            # requested_attributes requires pycups >= 1.9.42.
-                            attrs = c.getJobAttributes (jobid)
-
-                        jobdata.update (attrs)
-                    except cups.IPPError:
-                        pass
-                    except RuntimeError:
-                        pass
-
             may_be_problem = True
             if (jobdata['job-state'] == cups.IPP_JOB_HELD and
                 jobdata['job-hold-until'] == 'auth-info-required'):
-                # Leave this to update_job to deal with.
+                # update_job already dealt with this.
                 may_be_problem = False
             else:
                 # Other than that, unfortunately the only
@@ -1159,6 +1162,7 @@ class JobViewer (monitor.Watcher):
                                                       notify_text)
 
             if may_be_problem:
+                debugprint ("Problem detected")
                 self.toggle_window_display (self.statusicon, force_show=True)
                 dialog = gtk.Dialog (_("Print Error"), self.MainWindow, 0,
                                      (_("_Diagnose"), gtk.RESPONSE_NO,
@@ -1199,8 +1203,6 @@ class JobViewer (monitor.Watcher):
                 self.stopped_job_prompts.add (jobid)
                 dialog.show_all ()
 
-        self.update_job (jobid, jobdata)
-
     def job_removed (self, mon, jobid, eventname, event):
         monitor.Watcher.job_removed (self, mon, jobid, eventname, event)
         if self.jobiters.has_key (jobid):
@@ -1211,6 +1213,11 @@ class JobViewer (monitor.Watcher):
         if jobid in self.active_jobs:
             self.active_jobs.remove (jobid)
 
+        if self.auth_notifications.has_key (jobid):
+            self.auth_notifications[jobid].close ()
+            self.auth_notifications[jobid].set_data ('closed', True)
+            del self.auth_notifications[jobid]
+
         self.update_status ()
 
     def state_reason_added (self, mon, reason):
@@ -1278,7 +1285,10 @@ class JobViewer (monitor.Watcher):
         tuple = reason.get_tuple ()
         try:
             notification = self.state_reason_notifications[tuple]
-            notification.close ()
+            if notification.get_data ('closed') != True:
+                notification.close ()
+            del self.state_reason_notifications[tuple]
+            self.set_statusicon_visibility ()
         except KeyError:
             pass
 
@@ -1324,7 +1334,9 @@ class JobViewer (monitor.Watcher):
             debugprint ("Unexpected now_connected signal")
             return
 
-        notification.close ()
+        if notification.get_data ('closed') != True:
+            notification.close ()
+            notification.set_data ('closed', True)
 
     def printer_event (self, mon, printer, eventname, event):
         monitor.Watcher.printer_event (self, mon, printer, eventname, event)
